from pytorch_lightning.loggers import TestTubeLogger, NeptuneLogger
from pytorch_lightning.loggers.test_tube import Experiment
from pytorch_lightning.utilities import rank_zero_only
from pytorch_lightning.loggers.base import rank_zero_experiment
from typing import Optional
import os


class MMKExperiment(Experiment):
    """
    quick hack to get replace the versioning-subfolders of the original
    TestTube class with a unique "logs/" directory
    If you want to use experiments' names and versions by logging, just change the EXPERIMENT_CLASS
    attribute of MMKDefaultLogger to pl.loggers.test_tube.Experiment
    """

    def get_data_path(self, exp_name, exp_version):
        """
        Returns the path to the local package cache
        """
        if self.no_save_dir:
            return os.path.join(self.save_dir, 'test_tube_data', exp_name)
        else:
            return os.path.join(self.save_dir, exp_name)


class MMKDefaultLogger(TestTubeLogger):
    """
    quick hack to :
        - change the EXPERIMENT_CLASS inside a TestTubeLogger
        - format ALL hparams to strings for saving in logs/meta_tags.csv
          (otherwise some python objects get omitted...)
    """

    EXPERIMENT_CLASS = MMKExperiment

    def __init__(self, save_dir: str, version: Optional[int] = None):
        super(MMKDefaultLogger, self).__init__(save_dir, name="logs/", version=version)

    @property
    @rank_zero_experiment
    def experiment(self):
        if self._experiment is not None:
            return self._experiment

        self._experiment = self.EXPERIMENT_CLASS(
            save_dir=self.save_dir,
            name=self._name,
            debug=self.debug,
            version=self.version,
            description=self.description,
            create_git_tag=self.create_git_tag,
            rank=rank_zero_only.rank,
        )
        return self._experiment

    @rank_zero_only
    def log_hyperparams(self, params):
        super(MMKDefaultLogger, self).log_hyperparams({k: str(v) for k, v in params.items()})




from copy import copy

from ..data.data_object import DataObject


class DSWrapper(DataObject):

    def __init__(self):
        pass

    def upgrade(self, dataset):
        """
        dynamically extends an object with own methods and attributes.
        @param dataset: object to be extended
        @return: extended object will be an instance of object.__class__ and of self.__class__
        """
        new = copy(dataset)
        bases = (self.__class__, new.__class__)
        name = self.__class__.__name__ + "Dataset"
        new.__dict__.update(self.__dict__)
        new.__class__ = type(name, bases, new.__dict__)
        return new

    def __call__(self, dataset: DataObject):
        return self.upgrade(dataset)


class InputEqualTarget(DSWrapper):
    def __getitem__(self, item):
        return tuple(self.data[i] for i in [item, item])

    def __len__(self):
        return len(self.data)


class ShiftedSeqsPair(DSWrapper):
    def __init__(self, input_length, targets, stride=1):
        """
        @param input_length:
        @param targets: [(shift, length)... ]
        @param stride:
        """
        self.input_length = input_length
        self.shifts = list(zip(*targets))[0]
        self.lengths = list(zip(*targets))[1]
        self.stride = stride
        self.N = None

    def __call__(self, dataset: DataObject):
        # grab the number of time-steps BEFORE we upgrade
        self.N = len(dataset.data)
        return super(ShiftedSeqsPair, self).__call__(dataset)

    def __len__(self):
        ln = (self.N - max(self.lengths) - max(self.shifts) + 1) // self.stride
        return ln

    def __getitem__(self, item):
        i = item * self.stride
        input_slice = slice(i, i + self.input_length)
        target_slices = [slice(i + shift, i + shift + length)
                         for shift, length in zip(self.shifts, self.lengths)]
        inputs = self.data[input_slice]
        targets = tuple(self.data[idx] for idx in target_slices)
        return inputs, targets[0] if len(targets) == 1 else targets





import argparse
import os
from functools import partial
from itertools import tee
from multiprocessing import cpu_count

from mimikit.data import Database, make_root_db, file_to_fft, AudioFileWalker, upload_database

parser = argparse.ArgumentParser(prog="freqnet-db",
                                 description="transform audio file to FFTs with specified parameters and put "
                                             "them in .h5 Database that FreqNets can consume")

parser.add_argument("target", type=str,
                    help="the name (path) of the db you want to create")
parser.add_argument("--roots", "-r",
                    type=str, nargs="*",
                    default=None,
                    help="list of paths from which to search for audio files to include in the db")
parser.add_argument("--files", "-f",
                    type=str, nargs="*",
                    default=None,
                    help="list of audio files to include in the db")
parser.add_argument("--n-fft", '-d',
                    type=int, default=2048,
                    help="the fft size used to transform the files (default=2048)")
parser.add_argument("--hop-length", "-o",
                    type=int, default=512,
                    help="the hop length used to transform the files (default=512)")
parser.add_argument("--sample-rate", "-s",
                    type=int, default=22050,
                    help="the sample rate used to transform the files (default=22050)")
parser.add_argument("--neptune-project", '-p',
                    type=str, default=None,
                    help="name of the neptune.ai project you wish to upload the db to (requires that you stored your"
                         " neptune api token in the environment of this script)")


def freqnet_db(target,
               roots=None,
               files=None,
               n_fft=2048,
               hop_length=512,
               sample_rate=22050,
               neptune_project=None):
    namespace = argparse.Namespace(target=target, roots=roots, files=files, n_fft=n_fft,
                                   hop_length=hop_length, sample_rate=sample_rate,
                                   neptune_project=neptune_project)
    main(namespace)
    return Database(target)


def main(namespace=None):
    args = parser.parse_args(namespace=namespace)
    transform = partial(file_to_fft,
                        n_fft=args.n_fft,
                        hop_length=args.hop_length,
                        sr=args.sample_rate)
    if args.roots is None and args.files is None:
        args.roots = "./"
    walker = AudioFileWalker()
    walker, backup = tee(walker)
    print("Found following audio files :", "\n", *["\t" + file + "\n" for file in list(backup)])
    print("Making the database...")
    make_root_db(args.target, roots=args.roots, files=args.files, extract_func=transform, n_cores=cpu_count()//2)

    if args.neptune_project is not None:
        token = os.environ["NEPTUNE_API_TOKEN"]
        db = Database(args.target)
        print("uploading database to neptune...")
        upload_database(db, token, args.neptune_project, args.target)
    print("done!")


import torch

from .freqnet import FreqNet


class LayerWiseLossFreqNet(FreqNet):

    def forward(self, x):
        """
        """
        x = self.inpt(x)
        # we collect all the layers outputs
        skips = None
        outputs = []
        for layer in self.layers:
            x, skips = layer(x, skips)
            # decode each layer's output
            outputs += [self.outpt(skips)]
        if self.training:
            # pass them all to the loss function
            return tuple(outputs)
        # else redirect to a special method for inference
        return self.infer_layer_wise(outputs)

    def infer_layer_wise(self, outputs):
        """
        Place to implement the inference method when trained layer-wise.
        this default method returns the last output if the network isn't strict, otherwise, it collects the future
        time-steps where they are first predicted in the network (Xt+1 @ layer_1, Xt+2 @ layer_2...)
        @param outputs: list containing all the layers outputs
        @return: the prediction of the network
        """
        if not self.strict:
            return outputs[-1][-1]
        # we collect one future time step pro layer
        rf = self.receptive_field()
        future_steps = []
        for layer_out, shift in zip(outputs, self.all_shifts()):
            # we keep everything from the last layer, hence the condition on the slice
            step = slice(rf - shift, rf - shift +1) if rf - shift > 0 else slice(None, None)
            future_steps += [layer_out[:, step]]
        return torch.cat(future_steps, dim=1)

    def loss_fn(self, predictions, targets):
        denominator = len(self.layers)
        return sum(self._loss_fn(pred, trg) / denominator for pred, trg in zip(predictions, targets))

    def targets_shifts_and_lengths(self, input_length):
        shifts, lengths = self.all_shifts(), self.all_output_lengths(input_length)
        return list(zip(shifts, lengths))


from pytorch_lightning.callbacks.progress import Callback
from IPython import get_ipython


def is_notebook():
    shell = get_ipython().__class__.__name__
    if shell in ('ZMQInteractiveShell', "Shell"):
        # local and colab notebooks
        return True
    elif shell == 'TerminalInteractiveShell':
        return False
    else:
        return False


if is_notebook():
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm


class EpochProgressBarCallback(Callback):
    def on_train_start(self, trainer, pl_module):
        self.epoch_bar = tqdm(range(1, trainer.max_epochs), unit="epoch",
                              position=0, leave=False, dynamic_ncols=True)

    def on_epoch_end(self, trainer, pl_module):
        self.epoch_bar.update()




import torch
from pytorch_lightning import LightningModule, LightningDataModule
from torch.utils.data import DataLoader
import librosa
from abc import ABC

from ..data import DataObject, FeatureProxy, HOP_LENGTH
from ..kit import ShiftedSeqsPair, MMKHooks, LoggingHooks, tqdm


class ManyOneCycleLR(torch.optim.lr_scheduler.OneCycleLR):
    def step(self, epoch=None):
        try:
            super(ManyOneCycleLR, self).step(epoch=epoch)
        except ValueError:
            self.last_epoch = 0
            super(ManyOneCycleLR, self).step(epoch=epoch)


class FreqOptim:

    def __init__(self,
                 model,
                 max_lr=1e-3,
                 betas=(.9, .9),
                 div_factor=3.,
                 final_div_factor=1.,
                 pct_start=.25,
                 cycle_momentum=False):
        self.model = model
        self.max_lr = max_lr
        self.betas = betas
        self.div_factor = div_factor
        self.final_div_factor = final_div_factor
        self.pct_start = pct_start
        self.cycle_momentum = cycle_momentum
        # those are set in setup when we know what the trainer and datamodule got
        self.steps_per_epoch = None
        self.max_epochs = None

    def configure_optimizers(self):
        self.opt = torch.optim.Adam(self.model.parameters(), lr=self.max_lr, betas=self.betas)
        self.sched = ManyOneCycleLR(self.opt,
                                    steps_per_epoch=self.steps_per_epoch,
                                    epochs=self.max_epochs,
                                    max_lr=self.max_lr, div_factor=self.div_factor,
                                    final_div_factor=self.final_div_factor,
                                    pct_start=self.pct_start,
                                    cycle_momentum=self.cycle_momentum
                                    )
        return [self.opt], [{"scheduler": self.sched, "interval": "step", "frequency": 1}]


class FreqData(LightningDataModule):
    def __init__(self,
                 model,
                 data_object=None,
                 input_seq_length=64,
                 batch_size=64,
                 to_gpu=True,
                 splits=None,
                 **loader_kwargs,
                 ):
        super(FreqData, self).__init__()
        self.model = model
        self.ds = DataObject(data_object)
        self.input_seq_length = input_seq_length
        self.batch_size = batch_size
        self.to_gpu = to_gpu
        self.splits = splits
        loader_kwargs.setdefault("drop_last", False)
        loader_kwargs.setdefault("shuffle", True)
        self.loader_kwargs = loader_kwargs
        self.train_ds, self.val_ds, self.test_ds = None, None, None

    def prepare_data(self, *args, **kwargs):
        if not (getattr(self.model, "targets_shifts_and_lengths", False)):
            raise TypeError("Expected `model` to implement `targets_shifts_and_lengths(input_length)`"
                            " in order to compute the right slices for the batches")
        targets_def = self.model.targets_shifts_and_lengths(self.input_seq_length)
        wrapper = ShiftedSeqsPair(self.input_seq_length, targets_def)
        self.ds = wrapper(self.ds)

    def setup(self, stage=None):
        if stage == "fit":
            if self.to_gpu and torch.cuda.is_available():
                self.ds.to_tensor()
                self.ds.to("cuda")
            if self.splits is None:
                self.splits = (1.,)
            sets = self.ds.split(self.splits)
            for ds, attr in zip(sets, ["train_ds", "val_ds", "test_ds"]):
                setattr(self, attr, ds)

    def train_dataloader(self):
        if not self.has_prepared_data:
            self.prepare_data()
        if not self.has_setup_fit:
            self.setup("fit")
        return DataLoader(self.train_ds, batch_size=self.batch_size, **self.loader_kwargs)

    def val_dataloader(self, shuffle=False):
        has_val = len(self.splits) >= 2 and self.splits[1] is not None
        if not has_val:
            return None
        if not self.has_prepared_data:
            self.prepare_data()
        if not self.has_setup_fit:
            self.setup("fit")
        kwargs = self.loader_kwargs.copy()
        kwargs["shuffle"] = shuffle
        return DataLoader(self.val_ds, batch_size=self.batch_size, **kwargs)

    def test_dataloader(self, shuffle=False):
        has_test = len(self.splits) >= 3 and self.splits[2] is not None
        if not has_test:
            return None
        if not self.has_prepared_data:
            self.prepare_data()
        if not self.has_setup_fit:
            self.setup("test")
        kwargs = self.loader_kwargs.copy()
        kwargs["shuffle"] = shuffle
        return DataLoader(self.test_ds, batch_size=self.batch_size, **kwargs)


class FreqNetModel(MMKHooks,
                   LoggingHooks,
                   LightningModule,
                   ABC):

    def __init__(self,
                 data_object=None,
                 input_seq_length=64,
                 batch_size=64,
                 to_gpu=True,
                 splits=[.8, .2],
                 max_lr=1e-3,
                 betas=(.9, .9),
                 div_factor=3.,
                 final_div_factor=1.,
                 pct_start=.25,
                 cycle_momentum=False,
                 **loaders_kwargs):
        super(FreqNetModel, self).__init__()
        # dimensionality of inputs is automatically available
        if data_object is not None and not isinstance(data_object, str):
            self.input_dim = data_object.shape[-1]
            self.datamodule = FreqData(self, data_object, input_seq_length, batch_size,
                                       to_gpu, splits, **loaders_kwargs)
        else:
            raise ValueError("Please pass a valid data_object to instantiate a FreqNetModel."
                             " If you are loading from a checkpoint, you can do so by modifying your method call to :\n"
                             "FreqNetModel.load_from_checkpoint(path_to_ckpt, data_object=my_data_object)")
        self.optim = FreqOptim(self, max_lr, betas, div_factor, final_div_factor, pct_start,
                               cycle_momentum)
        # calling this updates self.hparams from any subclass : call it when subclassing!
        self.save_hyperparameters()

    def training_step(self, batch, batch_idx):
        batch, target = batch
        output = self.forward(batch)
        recon = self.loss_fn(output, target)
        return {"loss": recon}

    def validation_step(self, batch, batch_idx):
        batch, target = batch
        output = self.forward(batch)
        recon = self.loss_fn(output, target)
        return {"val_loss": recon}

    def setup(self, stage: str):
        if stage == "fit":
            self.logger.log_hyperparams(self.hparams)
            self.optim.max_epochs = self.trainer.max_epochs
            self.optim.steps_per_epoch = len(self.datamodule.train_dataloader())

    def configure_optimizers(self):
        return self.optim.configure_optimizers()

    def _set_hparams(self, hp):
        """
        Small handy hook to
            - copy params of a db in hparams
            - make sure we don't store arrays or tensors in hparams by replacing them
            by their string
        """
        # replace inputs passed to the hparams with their string
        if "data_object" in hp:
            if isinstance(hp["data_object"], FeatureProxy):
                # copy the attrs of the db to hparams
                for k, v in hp["data_object"].attrs.items():
                    hp[k] = v
            string = repr(hp["data_object"])
            hp["data_object"] = string[:88] + ("..." if len(string) > 88 else "")
        super(FreqNetModel, self)._set_hparams(hp)

    # Convenience Methods for generative audio models (see also, `LoggingHooks.log_audio`):

    def random_train_example(self):
        return next(iter(self.datamodule.train_dataloader()))[0][0:1]

    def random_val_example(self):
        return next(iter(self.datamodule.val_dataloader(shuffle=True)))[0][0:1]

    def generation_slices(self):
        raise NotImplementedError("subclasses of `FreqNetModel` have to implement `generation_slices`")

    def targets_shifts_and_lengths(self, input_length):
        raise NotImplementedError("subclasses of `FreqNetModel` have to implement `targets_shifts_and_lengths`")

    def generate(self, prompt, n_steps, hop_length=HOP_LENGTH, time_domain=True):
        was_training = self.training
        self.eval()
        initial_device = self.device
        self.to("cuda" if torch.cuda.is_available() else "cpu")
        if not isinstance(prompt, torch.Tensor):
            prompt = torch.from_numpy(prompt)
        if len(prompt.shape) < 3:
            prompt = prompt.unsqueeze(0)
        input_slice, output_slice = self.generation_slices()
        generated = prompt.to(self.device)
        for _ in tqdm(range(n_steps), desc="Generate", dynamic_ncols=True, leave=False, unit="step"):
            with torch.no_grad():
                out = self(generated[:, input_slice])
                generated = torch.cat((generated, out[:, output_slice]), dim=1)
        if time_domain:
            generated = generated.transpose(1, 2).squeeze()
            generated = librosa.griffinlim(generated.cpu().numpy(), hop_length=hop_length, n_iter=64)
            generated = torch.from_numpy(generated)
        self.to(initial_device)
        self.train() if was_training else None
        return generated.unsqueeze(0)



from torch import nn as nn


def mean_L1_prop(output, target):
    L = nn.L1Loss(reduction="none")(output, target).sum(dim=(0, -1), keepdim=True)
    return 100 * (L / target.abs().sum(dim=(0, -1), keepdim=True)).mean()


import torch
import torch.nn as nn

from .modules.loss_functions import mean_L1_prop
from .base import FreqNetModel
from .freqnet import FreqNet


class FrexpertMixture(FreqNetModel):
    """
    Use the Base FreqNet class to load a bunch of FreqNets for a mixture
    """

    def __init__(self, loss_fn=mean_L1_prop, **data_optim_kwargs):
        super(FrexpertMixture, self).__init__(**data_optim_kwargs)
        self.loss_fn = loss_fn
        self.save_hyperparameters()
        data_obj = data_optim_kwargs.get("data_object", None)
        # we use nets instead of layers, but we still store them in the layers attributes!
        self.nets = nn.ModuleList([
            FreqNet(data_object=data_obj,
                    model_dim=512,
                    groups=1,
                    n_layers=(3,),
                    strict=False,
                    accum_outputs=1),
            FreqNet(data_object=data_obj,
                    model_dim=1024,
                    groups=2,
                    n_layers=(3,),
                    strict=False,
                    accum_outputs=-1)
        ])
        # a simple MLP to output the weight of each net given the input
        self.proba_f = nn.Sequential(nn.Linear(1025, 1025), nn.Tanh(),
                                     nn.Linear(1025, 2), nn.Softmax(dim=-1))

    def forward(self, x):
        ys = torch.stack([net(x) for net in self.nets], dim=2)
        # we crop the begining of x to have matching shapes
        pr_ys = self.proba_f(x[:, -ys.size(1):]).unsqueeze(-1)
        # sum by the "net-dim"
        y = (pr_ys * ys).sum(dim=2)
        return y

    def targets_shifts_and_lengths(self, input_length):
        return self.nets[0].targets_shifts_and_lengths(input_length)


import torch.nn as nn


class GatedLinear(nn.Module):
    def __init__(self, in_dim, out_dim, **kwargs):
        super(GatedLinear, self).__init__()
        self.fcg = nn.Linear(in_dim, out_dim, **kwargs)
        self.fcf = nn.Linear(in_dim, out_dim, **kwargs)

    def forward(self, x):
        f = nn.Tanh()(self.fcf(x))
        g = nn.Sigmoid()(self.fcg(x))
        return f * g


class GatedConv(nn.Module):
    def __init__(self, c_in, c_out, kernel_size=2, transpose=False, **kwargs):
        super(GatedConv, self).__init__()
        mod = nn.Conv1d if not transpose else nn.ConvTranspose1d
        self.conv_f = mod(c_in, c_out, kernel_size, **kwargs)
        self.conv_g = mod(c_in, c_out, kernel_size, **kwargs)
        self.tanh = nn.Tanh()
        self.sig = nn.Sigmoid()

    def forward(self, x):
        f = self.tanh(self.conv_f(x))
        g = self.sig(self.conv_g(x))
        return f * g



from neptune import Session
from zipfile import ZipFile
import os
from getpass import getpass
import shutil


NEPTUNE_TOKEN_KEY = "NEPTUNE_API_TOKEN"


def get_token(key=NEPTUNE_TOKEN_KEY):
    api_token = os.environ.get(key, None)
    if api_token is None:
        api_token = getpass("Couldn't find your api token in environment. "
                            "Please paste your API token from neptune here: ")
        os.environ[key] = api_token
    return api_token


def get_neptune_experiment(api_token, full_exp_path: str):
    namespace, project, exp_id = full_exp_path.split("/")
    session = Session.with_default_backend(api_token=api_token)
    project = session.get_project(namespace + "/" + project)
    exps = project.get_experiments(id=exp_id)
    return None if not any(exps) else exps[0]


def get_neptune_project(api_token, project_name: str):
    namespace, project = project_name.split("/")
    session = Session.with_default_backend(api_token=api_token)
    project = session.get_project(namespace + "/" + project)
    return project


def download_model(api_token, full_exp_path, destination="./"):
    session = Session.with_default_backend(api_token=api_token)
    namespace, project, exp_id = full_exp_path.split("/")
    project_name = namespace + "/" + project
    model_project = session.get_project(project_name)
    exp = model_project.get_experiments(id=exp_id)[0]
    exp.download_artifacts('/', destination)
    with ZipFile(os.path.join(destination, "output.zip")) as f:
        f.extractall(destination)
    for subdir in os.listdir(os.path.join(destination, "output")):
        shutil.move(src=os.path.join(destination, "output", subdir),
                    dst=os.path.join(destination, subdir))
    os.remove(os.path.join(destination, "output.zip"))
    shutil.rmtree(os.path.join(destination, "output"))
    return 1



import pandas as pd
import numpy as np


def ssts(item, axis=0):
    """
    Start-Stop To Slices
    @param item: a pd.DataFrame with a "start" and a "stop" column (typically the metadata of a db)
    @param axis:
    @return: 1d array of slices for retrieving each element in `item`
    """
    arr = np.atleast_2d(item[['start', 'stop']].values)
    slices = map(lambda a: (*[slice(0, None)] * axis, slice(a[0], a[1]), ), arr)
    return slices


class Metadata(pd.DataFrame):

    # OVERRIDES

    @property
    def _constructor(self):
        return Metadata

    @staticmethod
    def _validate(obj):
        if 'start' not in obj.columns or "stop" not in obj.columns:
            raise ValueError("Must have 'start' and 'stop' columns.")

    # PROPERTIES

    @property
    def first_start(self):
        return self.start.min()

    @property
    def last_stop(self):
        return self.stop.max()

    @property
    def span(self):
        return self.last_stop - self.first_start

    @property
    def cumdur(self):
        return np.cumsum(self["duration"].values)

    @property
    def all_indices(self):
        return np.array([i for ev in self.events for i in range(ev.start, ev.stop)])

    def slices(self, time_axis=0):
        """
        This is the back-end core of a score. This method efficiently returns the indexing objects
        necessary to communicate with n-dimensional data.

        @return: an array of slices where slice_i corresponds to the row/Event_i in the DataFrame
        """
        return ssts(self, time_axis)

    @property
    def durations_(self):
        return self.stop.values - self.start.values

    @property
    def events(self):
        return self.itertuples(name="Event", index=True)

    # UPDATING METHOD

    def make_contiguous(self):
        self.reset_index(drop=True, inplace=True)
        cumdur = self.cumdur
        self.loc[:, "start"] = np.r_[0, cumdur[:-1]] if cumdur[0] != 0 else cumdur[:-1]
        self.loc[:, "stop"] = cumdur
        return self

    # Sanity Checks

    def is_consistent(self):
        return (self["duration"].values == (self.stop.values - self.start.values)).all()

    def is_contiguous(self):
        return (self.start.values[1:] == self.stop.values[:-1]).all()

    @staticmethod
    def from_start_stop(starts, stops, durations):
        return Metadata((starts, stops, durations), index=["start", "stop", "duration"]).T

    @staticmethod
    def from_stop(stop):
        """
        integers in `stops` correspond to the prev[stop] and next[start] values.
        `stops` must contain the last index ! and it can begin with 0, but doesn't have to...
        """
        stop = np.asarray(stop)
        if stop[0] == 0:
            starts = np.r_[0, stop[:-1]]
            stop = stop[1:]
        else:
            starts = np.r_[0, stop[:-1]]
        durations = stop - starts
        return Metadata.from_start_stop(starts, stop, durations)

    @staticmethod
    def from_duration(duration):
        stops = np.cumsum(duration)
        starts = np.r_[0, stops[:-1]]
        return Metadata.from_start_stop(starts, stops, duration)

    @staticmethod
    def from_data(sequence, time_axis=1):
        duration = np.array([x.shape[time_axis] for x in sequence])
        return Metadata.from_duration(duration)

    @staticmethod
    def from_frame_definition(total_duration, frame_length, stride=1, butlasts=0):
        starts = np.arange(total_duration - frame_length - butlasts + 1, step=stride)
        durations = frame_length + np.zeros_like(starts, dtype=np.int)
        stops = starts + durations
        return Metadata.from_start_stop(starts, stops, durations)


@pd.api.extensions.register_dataframe_accessor("soft_q")
class SoftQueryAccessor:
    def __init__(self, pandas_obj):
        self._df = pandas_obj

    def or_(self, **kwargs):
        series = False
        for col_name, func in kwargs.items():
            series = series | func(self._df[col_name])
        return series

    def and_(self, **kwargs):
        series = True
        for col_name, func in kwargs.items():
            series = series & func(self._df[col_name])
        return series


@pd.api.extensions.register_dataframe_accessor("hard_q")
class HardQueryAccessor:
    def __init__(self, pandas_obj):
        self._df = pandas_obj

    def or_(self, **kwargs):
        series = self._df.soft_q.or_(**kwargs)
        return self._df[series]

    def and_(self, **kwargs):
        series = self._df.soft_q.and_(**kwargs)
        return self._df[series]



import torch
import numpy as np
from torch.utils.data import Dataset as TorchDataset, Subset, random_split, DataLoader
from torch.utils.data._utils.collate import default_convert
from functools import update_wrapper


def map_if_multi(attr):
    def decorator(fn):
        def wrapper(self, *arg, **kwargs):
            if isinstance(getattr(self, attr), tuple):
                return tuple(getattr(x, fn.__name__)(*arg, **kwargs)[0] for x in getattr(self, attr))
            else:
                # for consistency, we always return a tuple
                # noinspection PyRedundantParentheses
                return (fn(self, *arg, **kwargs), )

        return update_wrapper(wrapper, fn,
                              # for some reason, __mod__ causes errors when using functools.wraps
                              ("__name__", '__qualname__', '__doc__', '__annotations__'),
                              ("__dict__",))
    return decorator


class DataObject(TorchDataset):

    @property
    def data(self):
        """
        give access to the underlying data to wrappers and co
        """
        return self._object

    @property
    def n_features(self):
        return self._n_features

    @property
    def dims(self):
        return self._shape

    @property
    def shape(self):
        return self._shape

    @property
    def dtype(self):
        return self._dtype

    @property
    def device(self):
        return self._device

    @property
    def style(self):
        return self._style

    @property
    def attrs(self):
        """
        @return: `attrs` : dict of user-specified infos he wants to keep close (e.g. {"sr": 22050, "time-axis": 0})
        """
        return self._attrs

    def __init__(self,
                 data_object=None,
                 wrappers=None,
                 augmentations=None,
                 attrs=None):
        """
        Here we initialize the instance.
        """
        if data_object is None:
            raise ValueError("`data_object` can not be None.")
        self._object = data_object
        self._wrapper = wrappers
        self._augmentation = augmentations
        self._attrs = attrs
        self._n_features = 1 if not isinstance(data_object, tuple) else len(data_object)
        self._is_multi = self._n_features > 1

        if self._is_multi:
            # we cast to Dataset recursively
            self._object = tuple(x if isinstance(x, DataObject) else DataObject(x) for x in self._object)

        self._style = self.get_style()
        if any(style is None for style in self._style):
            raise ValueError("Expected data_object to either implement __getitem__ and __len__, or __iter__."
                             " object of class %s implements none of those." % str(type(data_object)))
        if self._is_multi:
            if not all(self._style[0] == style for style in self._style[1:]):
                raise TypeError("Expected all data_objects to be of the same style. Got %s" % str(self._style))
            lengths = tuple(len(obj) for obj in self._object)
            if self._style[0] is "map" and not all(lengths[0] == n for n in lengths[1:]):
                raise ValueError("Expected all 'map-style' data_objects to be of same lengths. Got %s" % \
                                 str(lengths))
        self._dtype = self.get_dtype()
        if any(dtype is None for dtype in self._dtype):
            raise TypeError("Expected elements of `data_object` to be of torch, numpy or built-in numerical dtype."
                            " Got following dtypes %s" % str(self._dtype))
        self._shape = self.get_shape()
        self._device = self.get_device()

    def __call__(self, *args, **kwargs):
        """
        Method used to wrap around an other Dataset. Only used in the wrapper classes.
        """
        pass

    def __len__(self):
        return len(self._object[0]) if self._is_multi else len(self._object)

    @map_if_multi("_object")
    def __getitem__(self, item):
        if isinstance(item, tuple):
            return tuple(self._object[i] for i in item)
        return self._object[item]

    def __iter__(self):
        if isinstance(self._object, tuple):
            return zip(*[iter(obj) for obj in self._object])
        return iter(self._object)

    @staticmethod
    def has_len(data_object):
        return getattr(data_object, "__len__", False)

    @staticmethod
    def _get_style(data_object):
        if getattr(data_object, "__getitem__", False) and getattr(data_object, "__len__", False):
            return "map"
        elif getattr(data_object, "__iter__", False):
            return "iter"
        return None

    @map_if_multi("_object")
    def get_style(self):
        return DataObject._get_style(self._object)

    @staticmethod
    def _get_elem(data_object):
        if DataObject._get_style(data_object) == "map":
            return data_object[0]
        elif DataObject._get_style(data_object) == "iter":
            return next(iter(data_object))
        return None

    @map_if_multi("_object")
    def get_elem(self):
        return DataObject._get_elem(self._object)

    @staticmethod
    def _get_shape(data_object):
        shape = getattr(data_object, "shape", None)
        if shape is None:
            # we try to fet lengths recursively and add `None` when we get elements with no length
            # thus, in the case of a hierarchy of generators, returning at least the number of 'axes'
            N = len(data_object) if DataObject.has_len(data_object) else None
            shape = (N, )
            elem = DataObject._get_elem(data_object)
            while elem is not None and DataObject._get_elem(elem) != elem:
                shape = (*shape, len(elem) if DataObject.has_len(elem) else None)
                elem = DataObject._get_elem(elem)
            # last None in the shape is bottom-level element
            shape = shape[:-1]
        if isinstance(shape, torch.Size):
            shape = tuple(shape)
        return shape

    @map_if_multi("_object")
    def get_shape(self):
        return DataObject._get_shape(self._object)

    @staticmethod
    def _get_dtype(data_object):
        elem = DataObject._get_elem(data_object)
        dtype = getattr(data_object, "dtype", None)
        if dtype is None:
            dtype = getattr(elem, "dtype", None) or (type(elem) if type(elem) in (int, float, complex) else None)
            while dtype is None and elem is not None and DataObject._get_elem(elem) != elem:
                elem = DataObject._get_elem(elem)
                dtype = getattr(elem, "dtype", None) or (type(elem) if type(elem) in (int, float, complex) else None)
        if isinstance(dtype, torch.dtype):
            pass
        elif dtype in (float, int, complex):
            dtype = getattr(torch, np.dtype(dtype).name, None)
        elif getattr(dtype, "name", False):  # it's a numpy array!
            dtype = getattr(torch, dtype.name, dtype)
        return dtype

    @map_if_multi("_object")
    def get_dtype(self):
        return DataObject._get_dtype(self._object)

    @staticmethod
    def _get_device(data_object):
        elem = DataObject._get_elem(data_object)
        device = getattr(data_object, "device", None)
        if device is None:
            device = getattr(elem, "device", "cpu")
        if isinstance(device, torch.device):
            device = device.type
        return device

    @map_if_multi("_object")
    def get_device(self):
        return DataObject._get_device(self._object)

    @map_if_multi("_object")
    def to_tensor(self):
        if issubclass(type(self._object), torch.Tensor):
            return
        elif "iter" in self.style:
            raise TypeError("Cannot convert 'iter' style objects to tensor.")
        maybe_tensor = default_convert(self._object[:])
        if issubclass(type(maybe_tensor), torch.Tensor):
            self._object = maybe_tensor
            return
        try:
            self._object = torch.tensor(self._object)
        except Exception as e:
            raise e
        return

    @map_if_multi("_object")
    def select(self, indices, inplace=False):
        if not self.has_len(self._object):
            raise TypeError("Cannot select indices of objects that don't implement `__len__`")
        if inplace:
            self._object = self._object[np.sort(indices)]
        else:
            self._object = Subset(self._object, indices)

    @map_if_multi("_object")
    def to(self, device):
        """move any underlying tensor to some device"""
        if isinstance(self._object, torch.Tensor):
            self._object = self._object.to(device)

    def split(self, splits):
        """
        @param splits: Sequence of floats or ints possibly containing None. The sequence of elements
        corresponds to the proportion (floats), the number of examples (ints) or the absence of
        train-set, validation-set, test-set, other sets... in that order.
        @return: the *-sets
        """
        nones = []
        if any(x is None for x in splits):
            if splits[0] is None:
                raise ValueError("the train-set's split cannot be None")
            nones = [i for i, x in zip(range(len(splits)), splits) if x is None]
            splits = [x for x in splits if x is not None]
        if all(type(x) is float for x in splits):
            splits = [x / sum(splits) for x in splits]
            N = len(self)
            # leave the last one out for now because of rounding
            as_ints = [int(N * x) for x in splits[:-1]]
            # check that the last is not zero
            if N - sum(as_ints) == 0:
                raise ValueError("the last split rounded to zero element. Please provide a greater float or consider "
                                 "passing ints.")
            as_ints += [N - sum(as_ints)]
            splits = as_ints
        sets = list(random_split(self, splits))
        if any(nones):
            sets = [None if i in nones else sets.pop(0) for i in range(len(sets + nones))]
        return tuple(sets)

    def load(self, **kwargs):
        """pack self into a batch producer (Dataloader)"""
        return DataLoader(self, **kwargs)

    def __repr__(self):
        return "<DataObject shape:%s, dtype:%s>" % (self.shape, self.dtype)


from .callbacks import *
from .checkpoint import *
from .loggers import *
from .hooks import *
from .get_trainer import get_trainer
from .ds_wrappers import *
from .neptune_project import download_model, get_neptune_experiment, get_token, get_neptune_project




import librosa
import numpy as np
from librosa.display import specshow
import matplotlib.pyplot as plt
import IPython.display as ipd
import torch
from .data.transforms import HOP_LENGTH, SR

# Conversion

normalize = librosa.util.normalize
a2db = lambda S: librosa.amplitude_to_db(abs(S), ref=S.max())
s2f = librosa.samples_to_frames
s2t = librosa.samples_to_time
f2s = librosa.frames_to_samples
f2t = librosa.frames_to_time
t2f = librosa.time_to_frames
t2s = librosa.time_to_samples
hz2m = librosa.hz_to_midi
m2hz = librosa.midi_to_hz


# Debugging utils

def to_db(S):
    if S.dtype == np.complex64:
        S_hat = a2db(abs(S)) + 40
    elif S.min() >= 0 and S.dtype in (np.float, np.float32, np.float64, np.float_):
        S_hat = a2db(S) + 40
    else:
        S_hat = a2db(S) + 40
    return S_hat


def signal(S, hop_length=HOP_LENGTH):
    if S.dtype in (np.complex64, np.complex128):
        return librosa.istft(S, hop_length=hop_length)
    else:
        return librosa.griffinlim(S, hop_length=hop_length, n_iter=64)


def audio(S, hop_length=HOP_LENGTH, sr=SR):
    if len(S.shape) > 1:
        y = signal(S, hop_length)
        if y.size > 0:
            return ipd.display(ipd.Audio(y, rate=sr))
        else:
            return ipd.display(ipd.Audio(np.zeros(hop_length*2), rate=sr))
    else:
        return ipd.display(ipd.Audio(S, rate=sr))


def playlist(iterable):
    for seg in iterable:
        audio(seg)
    return


def playthrough(iterable, axis=1):
    rv = np.concatenate(iterable, axis=axis)
    return audio(rv)


def show(S, figsize=(), db_scale=True, title="", **kwargs):
    S_hat = to_db(S) if db_scale else S
    if figsize:
        plt.figure(figsize=figsize)
    if "x_axis" not in kwargs:
        kwargs["x_axis"] = "frames"
    if "y_axis" not in kwargs:
        kwargs["y_axis"] = "frames"
    ax = specshow(S_hat, sr=SR, **kwargs)
    plt.colorbar()
    plt.tight_layout()
    plt.title(title)
    return ax


# Array <==> Tensor ops :

def numcpu(y):
    if isinstance(y, torch.Tensor):
        if y.requires_grad:
            return y.detach().cpu().numpy()
        return y.cpu().numpy()
    else:  # tuples
        return tuple(numcpu(x) for x in y)


def cuda_if_available():
    return torch.device('cuda' if torch.cuda.is_available() else 'cpu')


def to_torch(x, device=cuda_if_available()):
    if isinstance(x, np.ndarray):
        return torch.from_numpy(x).float().to(device)
    elif isinstance(x, torch.Tensor):
        if x.device == device:
            return x.float()
        else:
            return x.float().to(device)
    else:  # tuples
        return tuple(to_torch(y) for y in x)



from .api import Database, FeatureProxy, download_database, upload_database
from .data_object import DataObject
from .factory import make_root_db, AudioFileWalker
from .metadata import Metadata
from .transforms import file_to_fft, default_extract_func, N_FFT, HOP_LENGTH, SR
